"""Bernardi-Raugel elements on simplices.

This element's definition appears in https://doi.org/10.2307/2007793
(Bernardi and Raugel, 1985)
"""

import sympy
from ..finite_element import CiarletElement
from ..moments import make_integral_moment_dofs
from ..polynomials import polynomial_set
from ..functionals import (NormalIntegralMoment, DotPointEvaluation,
                           DivergenceIntegralMoment)
from ..symbolic import x
from .lagrange import Lagrange, DiscontinuousLagrange


class BernardiRaugel(CiarletElement):
    """Bernardi-Raugel Hdiv finite element."""

    def __init__(self, reference, order):
        poly = polynomial_set(reference.tdim, reference.tdim, order)

        p = Lagrange(reference, 1, variant="equispaced")

        for i in range(reference.sub_entity_count(reference.tdim - 1)):
            sub_e = reference.sub_entity(reference.tdim - 1, i)
            bubble = 1
            for j in reference.sub_entities(reference.tdim - 1)[i]:
                bubble *= p.get_basis_function(j)
            poly.append(tuple(bubble * j for j in sub_e.normal()))

        dofs = []

        for n in range(reference.sub_entity_count(reference.tdim - 1)):
            facet = reference.sub_entity(reference.tdim - 1, n)
            for v in facet.vertices:
                dofs.append(DotPointEvaluation(
                    v, tuple(i * facet.jacobian() for i in facet.normal()),
                    entity=(reference.tdim - 1, n),
                    mapping="contravariant"))

        dofs += make_integral_moment_dofs(
            reference,
            facets=(NormalIntegralMoment, DiscontinuousLagrange, 0, "contravariant",
                    {"variant": "equispaced"}),
        )

        if order > 1:
            assert order == 2 and reference.name == "tetrahedron"

            for i in range(reference.tdim):
                poly.append(tuple(
                    p.get_basis_function(0) * p.get_basis_function(1)
                    * p.get_basis_function(2) * p.get_basis_function(3)
                    if j == i else 0 for j in range(reference.tdim)))

            for e_n, edge in enumerate(reference.edges):
                v1 = reference.vertices[edge[0]]
                v2 = reference.vertices[edge[1]]
                midpoint = tuple(sympy.Rational(i + j, 2) for i, j in zip(v1, v2))
                d = tuple(j - i for i, j in zip(v1, v2))
                dofs.append(DotPointEvaluation(midpoint, d, entity=(1, e_n),
                                               mapping="contravariant"))
            for f_n in range(reference.sub_entity_count(2)):
                face = reference.sub_entity(2, f_n)
                normal = [i * face.jacobian() for i in face.normal()]
                for e_n in range(3):
                    edge = face.sub_entity(1, e_n)
                    midpoint = tuple(sympy.Rational(i + j, 2) for i, j in zip(*edge.vertices))
                    dofs.append(DotPointEvaluation(midpoint, normal, entity=(2, f_n),
                                                   mapping="contravariant"))

            p = DiscontinuousLagrange(reference, 0, variant="equispaced")

            for i in range(3):
                d = tuple(1 if j == i else 0 for j in range(reference.tdim))
                dofs.append(DivergenceIntegralMoment(
                    reference, x[i], p.dofs[0], entity=(3, 0),
                    mapping="contravariant"
                ))

        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim)

    names = ["Bernardi-Raugel"]
    references = ["triangle", "tetrahedron"]
    min_order = 1
    max_order = {"triangle": 1, "tetrahedron": 2}
    continuity = "H(div)"



"""Polynomial sets."""
from .symbolic import x, Monomial
import sympy
import numpy


def get_index(refname, indices, max_order):
    """Get the index of a monomial."""
    if refname == "interval":
        return indices[0]
    if refname == "triangle":
        return sum(indices) * (sum(indices) + 1) // 2 + indices[1]
    if refname == "quadrilateral":
        return indices[1] * (max_order + 1) + indices[0]
    if refname == "tetrahedron":
        return (sum(indices) * (sum(indices) + 1) * (sum(indices) + 2) // 6
                + sum(indices[1:]) * (sum(indices[1:]) + 1) // 2 + indices[2])
    if refname == "hexahedron":
        return indices[2] * (max_order + 1) ** 2 + indices[1] * (max_order + 1) + indices[0]
    if refname == "prism":
        return ((max_order + 2) * (max_order + 1) * indices[2] // 2
                + sum(indices[:2]) * (sum(indices[:2]) + 1) // 2 + indices[1])


def num_polynomials(refname, max_order):
    """Get the number of polynomials."""
    if refname == "interval":
        return max_order + 1
    if refname == "triangle":
        return (max_order + 1) * (max_order + 2) // 2
    if refname == "quadrilateral":
        return (max_order + 1) ** 2
    if refname == "tetrahedron":
        return (max_order + 1) * (max_order + 2) * (max_order + 3) // 6
    if refname == "hexahedron":
        return (max_order + 1) ** 3
    if refname == "prism":
        return (max_order + 1) ** 2 * (max_order + 2) // 2


def get_max_order(basis, refname):
    """Get the maximum order of a basis."""
    if refname in ["interval", "triangle", "tetrahedron"]:
        return max(sum(i.indices) for i in basis)
    if refname in ["quadrilateral", "hexahedron"]:
        return max(max(i.indices) for i in basis)
    if refname == "prism":
        return max(max(sum(i.indices[:2]), i.indices[2]) for i in basis)


def get_min_order(basis, refname):
    """Get the minimum order of a basis."""
    if refname in ["interval", "triangle", "tetrahedron"]:
        return min(sum(i.indices) for i in basis)
    if refname in ["quadrilateral", "hexahedron"]:
        return min(max(i.indices) for i in basis)
    if refname == "prism":
        return min(max(sum(i.indices[:2]), i.indices[2]) for i in basis)


def _jrc(a, n, divide):
    """Coefficients in Jacobi polynomial recurrence relation."""
    return (
        divide((a + 2 * n + 1) * (a + 2 * n + 2), 2 * (n + 1) * (a + n + 1)),
        divide(a * a * (a + 2 * n + 1), 2 * (n + 1) * (a + n + 1) * (a + 2 * n)),
        divide(n * (a + n) * (a + 2 * n + 2), (n + 1) * (a + n + 1) * (a + 2 * n))
    )


def _legendre_interval(max_order, pts, leg, set_leg, divide):
    """Compute Legendre polynomials on an interval."""
    set_leg(0, 1)
    if max_order > 0:
        set_leg(1, 2 * pts(0) - 1)
    for n in range(1, max_order):
        set_leg(n + 1, ((2 * n + 1) * leg(1) * leg(n) - n * leg(n - 1)) / (n + 1))


def _legendre_quadrilateral(max_order, pts, leg, set_leg, divide):
    """Compute Legendre polynomials on a quadrilateral."""

    def ind(a, b):
        return get_index("quadrilateral", (a, b), max_order)

    def pts1(i):
        return pts(1 + i)

    def leg1(i):
        return leg(ind(0, i))

    def set_leg1(i, value):
        set_leg(ind(0, i), value)

    _legendre_interval(max_order, pts, leg, set_leg, divide)
    _legendre_interval(max_order, pts1, leg1, set_leg1, divide)

    for i in range(1, max_order + 1):
        for j in range(1, max_order + 1):
            set_leg(ind(i, j), leg(ind(i, 0)) * leg(ind(0, j)))


def _legendre_hexahedron(max_order, pts, leg, set_leg, divide):
    """Compute Legendre polynomials on a hexahedron."""

    def ind(a, b, c):
        return get_index("hexahedron", (a, b, c), max_order)

    def pts1(i):
        return pts(1 + i)

    def leg1(i):
        return leg(ind(0, i, 0))

    def set_leg1(i, value):
        set_leg(ind(0, i, 0), value)

    def pts2(i):
        return pts(2 + i)

    def leg2(i):
        return leg(ind(0, 0, i))

    def set_leg2(i, value):
        set_leg(ind(0, 0, i), value)

    _legendre_interval(max_order, pts, leg, set_leg, divide)
    _legendre_interval(max_order, pts1, leg1, set_leg1, divide)
    _legendre_interval(max_order, pts2, leg2, set_leg2, divide)

    for i in range(max_order + 1):
        for j in range(max_order + 1):
            for k in range(max_order + 1):
                set_leg(ind(i, j, k), leg(ind(i, 0, 0)) * leg(ind(0, j, 0)) * leg(ind(0, 0, k)))


def _legendre_prism(max_order, pts, leg, set_leg, divide):
    """Compute Legendre polynomials on a prism."""

    def ind(a, b, c):
        return get_index("prism", (a, b, c), max_order)

    def pts2(i):
        return pts(2 + i)

    def leg2(i):
        return leg(ind(0, 0, i))

    def set_leg2(i, value):
        set_leg(ind(0, 0, i), value)

    _legendre_triangle(max_order, pts, leg, set_leg, divide)
    _legendre_interval(max_order, pts2, leg2, set_leg2, divide)

    for i in range(max_order + 1):
        for j in range(max_order - i + 1):
            for k in range(max_order + 1):
                set_leg(ind(i, j, k), leg(ind(i, j, 0)) * leg(ind(0, 0, k)))


def _legendre_triangle(max_order, pts, leg, set_leg, divide):
    """Compute Legendre polynomials on a triangle."""

    def ind(a, b):
        return get_index("triangle", (a, b), max_order)

    set_leg(0, 1)

    for p in range(1, max_order + 1):
        a = divide(2 * p - 1, p)
        if p > 1:
            set_leg(ind(p, 0),
                    (pts(0) * 2 + pts(1) - 1) * leg(ind(p - 1, 0)) * a
                    - (1 - pts(1)) ** 2 * leg(ind(p - 2, 0)) * (a - 1))
        else:
            set_leg(ind(p, 0), (pts(0) * 2 + pts(1) - 1) * leg(ind(p - 1, 0)) * a)

    for p in range(max_order):
        set_leg(ind(p, 1),
                leg(ind(p, 0)) * ((pts(1) * 2 - 1) * (divide(3, 2) + p) + divide(1, 2) + p))
        for q in range(1, max_order - p):
            a1, a2, a3 = _jrc(2 * p + 1, q, divide)
            set_leg(ind(p, q + 1),
                    leg(ind(p, q)) * ((pts(1) * 2 - 1) * a1 + a2) - leg(ind(p, q - 1)) * a3)


def _legendre_tetrahedron(max_order, pts, leg, set_leg, divide):
    """Compute Legendre polynomials on a tetrahedron."""

    def ind(a, b, c):
        return get_index("tetrahedron", (a, b, c), max_order)

    set_leg(0, 1)

    for p in range(1, max_order + 1):
        a = divide(2 * p - 1, p)
        if p > 1:
            set_leg(ind(p, 0, 0),
                    (pts(0) * 2 + pts(1) + pts(2) - 1) * leg(ind(p - 1, 0, 0)) * a
                    - (pts(1) + pts(2) - 1) ** 2 * leg(ind(p - 2, 0, 0)) * (a - 1))
        else:
            set_leg(ind(p, 0, 0),
                    (pts(0) * 2 + pts(1) + pts(2) - 1) * leg(ind(p - 1, 0, 0)) * a)

    for p in range(max_order):
        set_leg(ind(p, 1, 0),
                leg(ind(p, 0, 0)) * (pts(1) * (3 + 2 * p) + pts(2) - 1))

        for q in range(1, max_order - p):
            a1, a2, a3 = _jrc(2 * p + 1, q, divide)
            set_leg(ind(p, q + 1, 0),
                    leg(ind(p, q, 0)) * (pts(1) * 2 * a1 + (pts(2) - 1) * (a1 - a2))
                    - leg(ind(p, q - 1, 0)) * (1 - pts(2)) ** 2 * a3)

    for p in range(max_order):
        for q in range(max_order - p):
            set_leg(ind(p, q, 1),
                    leg(ind(p, q, 0)) * ((1 + p + q) + (pts(2) * 2 - 1) * (2 + p + q)))

    for p in range(max_order - 1):
        for q in range(max_order - p - 1):
            for r in range(1, max_order - p - q):
                a1, a2, a3 = _jrc(2 * p + 2 * q + 2, r, divide)
                set_leg(ind(p, q, r + 1),
                        leg(ind(p, q, r)) * ((pts(2) * 2 - 1) * a1 + a2)
                        - leg(ind(p, q, r - 1)) * a3)


def evaluate_legendre_basis(points, basis, reference):
    """Evaluate the Legendre basis spanning the same set as the given basis."""
    for i in basis:
        if not isinstance(i, Monomial):
            return None
    max_order = get_max_order(basis, reference.name)
    min_order = get_min_order(basis, reference.name)

    if min_order is None or max_order is None or min_order < 0:
        return None

    num_p = num_polynomials(reference.name, max_order)
    if num_p is None:
        return None

    legendre = numpy.empty((len(points), num_polynomials(reference.name, max_order)))

    def pts(i):
        return points[:, i]

    def leg(i):
        return legendre[:, i]

    def set_leg(i, value):
        legendre[:, i] = value

    def divide(a, b):
        return a / b

    if reference.name == "interval":
        _legendre_interval(max_order, pts, leg, set_leg, divide)
    elif reference.name == "triangle":
        _legendre_triangle(max_order, pts, leg, set_leg, divide)
    elif reference.name == "tetrahedron":
        _legendre_tetrahedron(max_order, pts, leg, set_leg, divide)
    elif reference.name == "quadrilateral":
        _legendre_quadrilateral(max_order, pts, leg, set_leg, divide)
    elif reference.name == "hexahedron":
        _legendre_hexahedron(max_order, pts, leg, set_leg, divide)
    elif reference.name == "prism":
        _legendre_prism(max_order, pts, leg, set_leg, divide)

    if len(basis) == legendre.shape[1]:
        return legendre

    polys = numpy.empty((len(points), len(basis)))
    for i, b in enumerate(basis):
        polys[:, i] = legendre[:, get_index(reference.name, b.indices, max_order)]

    return polys


def get_legendre_basis(basis, reference):
    """Get the symbolic Legendre basis spanning the same set as the given basis."""
    for i, j in enumerate(basis):
        if isinstance(j, int):
            basis[i] = Monomial()
        elif not isinstance(j, Monomial):
            return None

    max_order = get_max_order(basis, reference.name)
    min_order = get_min_order(basis, reference.name)

    if min_order is None or max_order is None or min_order < 0:
        return None

    num_p = num_polynomials(reference.name, max_order)
    if num_p is None:
        return None

    legendre = [0 for i in range(num_polynomials(reference.name, max_order))]

    def pts(i):
        return x[i]

    def leg(i):
        return legendre[i]

    def set_leg(i, value):
        legendre[i] = value

    def divide(a, b):
        return sympy.Rational(a, b)

    if reference.name == "interval":
        _legendre_interval(max_order, pts, leg, set_leg, divide)
    elif reference.name == "triangle":
        _legendre_triangle(max_order, pts, leg, set_leg, divide)
    elif reference.name == "tetrahedron":
        _legendre_tetrahedron(max_order, pts, leg, set_leg, divide)
    elif reference.name == "quadrilateral":
        _legendre_quadrilateral(max_order, pts, leg, set_leg, divide)
    elif reference.name == "hexahedron":
        _legendre_hexahedron(max_order, pts, leg, set_leg, divide)
    elif reference.name == "prism":
        _legendre_prism(max_order, pts, leg, set_leg, divide)

    if len(basis) == len(legendre):
        return legendre

    polys = []
    for i, b in enumerate(basis):
        polys.append(legendre[get_index(reference.name, b.indices, max_order)])

    return polys



"""Functions to handle derivatives."""

from .vectors import vdot
from .symbolic import x, sym_sum, to_sympy


def derivative(f, dir):
    """Find the directional derivative of a function."""
    return vdot(grad(f, len(dir)), dir)


def grad(f, dim, variables=x):
    """Find the gradient of a scalar function."""
    return tuple(diff(f, variables[i]) for i in range(dim))


def jacobian_component(f, component):
    """Find a component of the Jacobian."""
    return diff(f, x[component[0]], x[component[1]])


def jacobian(f, dim):
    """Find the Jacobian."""
    return [[diff(f, x[i], x[j]) for i in range(dim)] for j in range(dim)]


def div(f):
    """Find the divergence of a vector function."""
    return sym_sum(diff(j, x[i]) for i, j in enumerate(f))


def curl(f):
    """Find the curl of a 3D vector function."""
    return (
        diff(f[2], x[1]) - diff(f[1], x[2]),
        diff(f[0], x[2]) - diff(f[2], x[0]),
        diff(f[1], x[0]) - diff(f[0], x[1])
    )


def diff(f, *vars):
    """Calculate the derivative of a function."""
    if isinstance(f, list):
        return [diff(i, *vars) for i in f]
    if isinstance(f, tuple):
        return tuple(diff(i, *vars) for i in f)

    out = to_sympy(f)
    for i in vars:
        out = out.diff(to_sympy(i))
    return out



"""Utility functions for testing."""

import sympy

test_elements = {
    "interval": {
        "P": [[{"variant": "equispaced"}, range(6)], [{"variant": "lobatto"}, range(4)]],
        "dP": [[{"variant": "equispaced"}, range(6)], [{"variant": "lobatto"}, range(4)],
               [{"variant": "radau"}, range(3)], [{"variant": "legendre"}, range(3)]],
        "vdP": [[{"variant": "equispaced"}, range(6)], [{"variant": "lobatto"}, range(3)],
                [{"variant": "radau"}, range(3)],
                [{"variant": "legendre"}, range(3)]],
        "vP": [[{"variant": "equispaced"}, range(6)], [{"variant": "lobatto"}, range(3)],
               [{"variant": "radau"}, range(3)], [{"variant": "legendre"}, range(3)]],
        "bubble": [[{"variant": "equispaced"}, range(2, 6)],
                   [{"variant": "lobatto"}, range(2, 3)]],
        "serendipity": [[{"variant": "equispaced"}, range(1, 6)],
                        [{"variant": "lobatto"}, range(1, 3)]],
        "Hermite": [[{}, [3]]],
        "Bernstein": [[{}, range(1, 4)]],
        "Taylor": [[{}, range(0, 5)]],
        "Wu-Xu": [[{}, [2]]],
        "MWX": [[{}, [1]]],
    },
    "triangle": {
        "P": [[{"variant": "equispaced"}, range(5)], [{"variant": "lobatto"}, range(3)]],
        "dP": [[{"variant": "equispaced"}, range(5)], [{"variant": "lobatto"}, range(3)]],
        "vdP": [[{"variant": "equispaced"}, range(5)], [{"variant": "lobatto"}, range(3)]],
        "vP": [[{"variant": "equispaced"}, range(5)], [{"variant": "lobatto"}, range(3)]],
        "matrix discontinuous Lagrange": [[{"variant": "equispaced"}, range(3)],
                                          [{"variant": "lobatto"}, range(3)]],
        "symmetric matrix discontinuous Lagrange": [
            [{"variant": "equispaced"}, range(3)], [{"variant": "lobatto"}, range(3)]],
        "bubble": [[{"variant": "equispaced"}, range(3, 5)], [{"variant": "lobatto"}, range(3, 3)]],
        "bubble enriched Lagrange": [[{"variant": "equispaced"}, range(1, 3)],
                                     [{"variant": "lobatto"}, range(1, 3)]],
        "bubble enriched vector Lagrange":
            [[{"variant": "equispaced"}, range(1, 3)], [{"variant": "lobatto"}, range(1, 3)]],
        "CR": [[{"variant": "equispaced"}, [1, 3, 5]]],
        "conforming CR": [[{}, range(1, 6)]],
        "HHJ": [[{"variant": "equispaced"}, range(3)]],
        "Bell": [[{"variant": "equispaced"}, [5]]],
        "Morley": [[{}, [2]]],
        "MWX": [[{}, [1, 2]]],
        "Regge": [[{}, range(4)]],
        "AW": [[{"variant": "equispaced"}, range(3, 5)]],
        "Nedelec1": [[{"variant": "equispaced"}, range(1, 4)]],
        "Nedelec2": [[{"variant": "equispaced"}, range(1, 4)]],
        "RT": [[{"variant": "equispaced"}, range(1, 4)]],
        "BDFM": [[{"variant": "equispaced"}, range(1, 4)]],
        "Argyris": [[{}, [5]]],
        "MTW": [[{"variant": "equispaced"}, [3]]],
        "KMV": [[{}, [1, 3]]],
        "Hermite": [[{}, [3]]],
        "BDM": [[{"variant": "equispaced"}, range(1, 4)]],
        "Bernstein": [[{}, range(1, 4)]],
        "HCT": [[{}, [3]]],
        "rHCT": [[{}, [3]]],
        "FS": [[{}, [2]]],
        "Taylor": [[{}, range(0, 5)]],
        "Bernardi-Raugel": [[{}, [1]]],
        "Wu-Xu": [[{}, [3]]],
        "transition": [[{"edge_orders": [1, 1, 1], "variant": "equispaced"}, range(1, 6)],
                       [{"edge_orders": [2, 2, 1], "variant": "equispaced"}, range(1, 6)],
                       [{"edge_orders": [3, 3, 1], "variant": "equispaced"}, range(1, 6)],
                       [{"edge_orders": [3, 3, 2], "variant": "equispaced"}, range(1, 6)],
                       [{"edge_orders": [2, 2, 1], "variant": "lobatto"}, range(1, 3)]],
    },
    "tetrahedron": {
        "P": [[{"variant": "equispaced"}, range(3)], [{"variant": "lobatto"}, range(3)]],
        "dP": [[{"variant": "equispaced"}, range(3)], [{"variant": "lobatto"}, range(3)]],
        "vdP": [[{"variant": "equispaced"}, range(3)], [{"variant": "lobatto"}, range(3)]],
        "vP": [[{"variant": "equispaced"}, range(3)], [{"variant": "lobatto"}, range(3)]],
        "matrix discontinuous Lagrange": [[{"variant": "equispaced"}, range(2)],
                                          [{"variant": "lobatto"}, range(3)]],
        "symmetric matrix discontinuous Lagrange": [[{"variant": "equispaced"}, range(2)],
                                                    [{"variant": "lobatto"}, range(3)]],
        "bubble": [[{"variant": "equispaced"}, [4]], [{"variant": "lobatto"}, [4]]],
        "CR": [[{"variant": "equispaced"}, [1]]],
        "Regge": [[{}, range(3)]],
        "Nedelec1": [[{"variant": "equispaced"}, range(1, 3)]],
        "Nedelec2": [[{"variant": "equispaced"}, range(1, 3)]],
        "RT": [[{"variant": "equispaced"}, range(1, 3)]],
        "BDFM": [[{"variant": "equispaced"}, range(1, 3)]],
        "MTW": [[{"variant": "equispaced"}, [3]]],
        "MWX": [[{}, [1, 2, 3]]],
        "KMV": [[{}, [1]]],
        "Hermite": [[{}, [3]]],
        "BDM": [[{"variant": "equispaced"}, range(1, 3)]],
        "Bernstein": [[{}, range(1, 3)]],
        "Taylor": [[{}, range(0, 5)]],
        "Bernardi-Raugel": [[{}, [1, 2]]],
        "Wu-Xu": [[{}, [4]]],
        "transition": [[{"edge_orders": [1, 1, 1, 1, 1, 1], "face_orders": [1, 1, 1, 1],
                         "variant": "equispaced"}, range(1, 5)],
                       [{"edge_orders": [1, 1, 1, 1, 1, 1], "face_orders": [3, 3, 1, 1],
                         "variant": "equispaced"}, range(1, 5)],
                       [{"edge_orders": [1, 1, 1, 1, 1, 1], "face_orders": [3, 3, 4, 2],
                         "variant": "equispaced"}, range(1, 5)],
                       [{"edge_orders": [1, 1, 1, 1, 1, 1], "face_orders": [4, 4, 1, 1],
                         "variant": "equispaced"}, range(1, 5)],
                       [{"edge_orders": [2, 2, 2, 2, 2, 2], "face_orders": [1, 1, 1, 1],
                         "variant": "equispaced"}, range(1, 5)],
                       [{"edge_orders": [3, 2, 1, 2, 1, 1], "face_orders": [1, 1, 1, 1],
                         "variant": "equispaced"}, range(1, 5)],
                       [{"edge_orders": [1, 1, 3, 1, 3, 1], "face_orders": [1, 1, 1, 1],
                         "variant": "equispaced"}, range(1, 5)],
                       [{"edge_orders": [1, 1, 2, 1, 2, 5], "face_orders": [3, 3, 2, 5],
                         "variant": "equispaced"}, range(1, 5)]],
    },
    "quadrilateral": {
        "dP": [[{"variant": "equispaced"}, range(4)], [{"variant": "lobatto"}, range(3)],
               [{"variant": "radau"}, range(3)], [{"variant": "legendre"}, range(3)]],
        "vdP": [[{"variant": "equispaced"}, range(4)], [{"variant": "lobatto"}, range(3)],
                [{"variant": "radau"}, range(3)], [{"variant": "legendre"}, range(3)]],
        "symmetric matrix discontinuous Lagrange": [
            [{"variant": "equispaced"}, range(2)]],
        "matrix discontinuous Lagrange": [
            [{"variant": "equispaced"}, range(2)]],
        "bubble": [[{"variant": "equispaced"}, range(2, 4)], [{"variant": "lobatto"}, range(2, 4)]],
        "Q": [[{"variant": "equispaced"}, range(4)], [{"variant": "lobatto"}, range(3)]],
        "dQ": [[{"variant": "equispaced"}, range(4)], [{"variant": "lobatto"}, range(3)]],
        "vQ": [[{"variant": "equispaced"}, range(4)], [{"variant": "lobatto"}, range(3)]],
        "serendipity": [[{"variant": "equispaced"}, range(1, 4)],
                        [{"variant": "lobatto"}, range(1, 3)]],
        "direct serendipity": [[{}, range(1, 7)]],
        "Scurl": [[{"variant": "equispaced"}, range(1, 4)], [{"variant": "lobatto"}, range(1, 3)]],
        "Sdiv": [[{"variant": "equispaced"}, range(1, 4)], [{"variant": "lobatto"}, range(1, 3)]],
        "Qcurl": [[{"variant": "equispaced"}, range(1, 4)]],
        "Qdiv": [[{"variant": "equispaced"}, range(1, 4)]],
        "BFS": [[{}, [3]]],
        "BDFM": [[{"variant": "equispaced"}, range(1, 4)]],
    },
    "hexahedron": {
        "dP": [[{"variant": "equispaced"}, range(3)], [{"variant": "lobatto"}, range(3)],
               [{"variant": "radau"}, range(3)], [{"variant": "legendre"}, range(3)]],
        "vdP": [[{"variant": "equispaced"}, range(3)], [{"variant": "lobatto"}, range(3)],
                [{"variant": "radau"}, range(3)], [{"variant": "legendre"}, range(3)]],
        "bubble": [[{"variant": "equispaced"}, range(2, 4)], [{"variant": "lobatto"}, range(2, 4)]],
        "symmetric matrix discontinuous Lagrange": [
            [{"variant": "equispaced"}, range(2)]],
        "matrix discontinuous Lagrange": [
            [{"variant": "equispaced"}, range(2)]],
        "Q": [[{"variant": "equispaced"}, range(3)], [{"variant": "lobatto"}, range(3)]],
        "dQ": [[{"variant": "equispaced"}, range(3)], [{"variant": "lobatto"}, range(3)]],
        "vQ": [[{"variant": "equispaced"}, range(3)], [{"variant": "lobatto"}, range(3)]],
        "serendipity": [[{"variant": "equispaced"}, range(1, 3)],
                        [{"variant": "lobatto"}, range(1, 3)]],
        "Scurl": [[{"variant": "equispaced"}, range(1, 3)], [{"variant": "lobatto"}, range(1, 3)]],
        "Sdiv": [[{"variant": "equispaced"}, range(1, 3)], [{"variant": "lobatto"}, range(1, 3)]],
        "Qcurl": [[{"variant": "equispaced"}, range(1, 3)]],
        "Qdiv": [[{"variant": "equispaced"}, range(1, 3)]],
        "BDFM": [[{"variant": "equispaced"}, range(1, 3)]],
        "BDDF": [[{"variant": "equispaced"}, range(1, 3)]],
    },
    "prism": {
        "Lagrange": [[{"variant": "equispaced"}, range(4)]],
        "Nedelec": [[{"variant": "equispaced"}, range(1, 3)]],
    },
    "pyramid": {
        "Lagrange": [[{"variant": "equispaced"}, range(4)]],
    }
}


def all_symequal(a, b):
    """Check if two symbolic numbers or vectors are equal."""
    if isinstance(a, (list, tuple)):
        for i, j in zip(a, b):
            if not all_symequal(i, j):
                return False
        return True
    return sympy.expand(sympy.simplify(a)) == sympy.expand(sympy.simplify(b))



import pytest
import numpy as np
from symfem import legendre, create_element
from symfem.symbolic import x, t, subs, to_float


def make_lattice(cell, N=3):
    import numpy as np
    if cell == "interval":
        return np.array([[i / N] for i in range(N + 1)])
    if cell == "triangle":
        return np.array([[i / N, j / N] for i in range(N + 1) for j in range(N + 1 - i)])
    if cell == "tetrahedron":
        return np.array([[i / N, j / N, k / N]
                         for i in range(N + 1) for j in range(N + 1 - i)
                         for k in range(N + 1 - i - j)])
    if cell == "quadrilateral":
        return np.array([[i / N, j / N] for i in range(N + 1) for j in range(N + 1)])
    if cell == "hexahedron":
        return np.array([[i / N, j / N, k / N]
                         for i in range(N + 1) for j in range(N + 1) for k in range(N + 1)])
    if cell == "prism":
        return np.array([[i / N, j / N, k / N]
                         for i in range(N + 1) for j in range(N + 1 - i) for k in range(N + 1)])
    if cell == "pyramid":
        return np.array([[i / N, j / N, k / N]
                         for i in range(N + 1) for j in range(N + 1)
                         for k in range(N + 1 - max(i, j))])


max_orders = {"interval": 10, "triangle": 8, "tetrahedron": 5, "quadrilateral": 8, "hexahedron": 5,
              "prism": 5, "pyramid": 5}


@pytest.mark.parametrize(
    ("cell", "order"),
    [(c, o) for c in ["interval", "triangle", "tetrahedron", "quadrilateral", "hexahedron",
                      "prism", "pyramid"] for o in range(1, max_orders[c])])
def test_legendre(cell, order):
    e = create_element(cell, "Lagrange", order)
    points = make_lattice(cell)

    basis = legendre.get_legendre_basis(e._basis, e.reference)
    values = legendre.evaluate_legendre_basis(points, e._basis, e.reference)

    if basis is None:
        assert values is None
        pytest.skip()

    for b in basis:
        assert b != 0

    values2 = np.array([[to_float(subs(b, x, p)) for b in basis] for p in points])

    assert np.allclose(values, values2)


@pytest.mark.parametrize("cell", ["interval", "triangle", "tetrahedron", "quadrilateral",
                                  "hexahedron", "prism", "pyramid"])
def test_orthogonal(cell):
    if cell == "interval":
        e = create_element(cell, "Lagrange", 5)
    else:
        e = create_element(cell, "Lagrange", 2)
    basis = legendre.get_legendre_basis(e._basis, e.reference)

    if basis is None:
        pytest.skip()

    for i, f in enumerate(basis):
        for g in basis[:i]:
            assert e.reference.integral(subs(f * g, x, t)) == 0



"""Lagrange elements on simplices."""

from itertools import product
from ..finite_element import CiarletElement
from ..polynomials import polynomial_set
from ..functionals import PointEvaluation
from ..quadrature import get_quadrature
from ..symbolic import x, subs
from .lagrange import DiscontinuousLagrange


class Transition(CiarletElement):
    """Transition finite element."""

    def __init__(self, reference, order,
                 edge_orders=None, face_orders=None, variant="equispaced"):
        if reference.name == "triangle":
            assert face_orders is None
            assert len(edge_orders) == 3
        elif reference.name == "tetrahedron":
            assert len(face_orders) == 4
            assert len(edge_orders) == 6

        bubble_space = DiscontinuousLagrange(reference, 1)

        dofs = []
        poly = polynomial_set(reference.tdim, 1, 1)
        for v_n, v in enumerate(reference.reference_vertices):
            dofs.append(PointEvaluation(v, entity=(0, v_n)))

        for edim in range(1, 4):
            for e_n in range(reference.sub_entity_count(edim)):
                entity = reference.sub_entity(edim, e_n)
                if edim == reference.tdim:
                    entity_order = order
                elif edim == 1:
                    entity_order = edge_orders[e_n]
                elif edim == 2:
                    entity_order = face_orders[e_n]
                else:
                    raise RuntimeError("Could not find order for this entity.")

                # DOFs
                points, _ = get_quadrature(variant, entity_order + 1)
                for i in product(range(1, entity_order), repeat=edim):
                    if sum(i) < entity_order:
                        dofs.append(
                            PointEvaluation(
                                tuple(o + sum(a[j] * points[b]
                                              for a, b in zip(entity.axes, i))
                                      for j, o in enumerate(entity.origin)),
                                entity=(edim, e_n)))

                # Basis
                if entity_order > edim:
                    if edim == reference.tdim:
                        bubble = 1
                        for f in bubble_space.get_basis_functions():
                            bubble *= f
                    elif edim == reference.tdim - 1:
                        bubble = 1
                        for i, f in enumerate(bubble_space.get_basis_functions()):
                            if i != e_n:
                                bubble *= f
                    else:
                        assert edim == 1 and reference.tdim == 3
                        bubble = 1
                        for i, f in enumerate(bubble_space.get_basis_functions()):
                            if i in reference.edges[e_n]:
                                bubble *= f
                    space = DiscontinuousLagrange(entity, entity_order - edim - 1,
                                                  variant=variant)
                    vars = []
                    origin = entity.vertices[0]
                    used = []
                    for p in entity.vertices[1:]:
                        i = 0
                        while p[i] == origin[i] or origin[i] == 1 or i in used:
                            i += 1
                        used.append(i)
                        vars.append(origin[i] + (p[i] - origin[i]) * x[i])

                    poly += [subs(f, x, vars) * bubble for f in space.get_basis_functions()]

        super().__init__(
            reference, order, poly, dofs, reference.tdim, 1
        )

    names = ["transition"]
    references = ["triangle", "tetrahedron"]
    min_order = 1
    continuity = "C0"



import symfem
from io import StringIO
import json
import sys
import os
import re
import pytest

code = False
output = False
lines = []
doc_data = []
outputlines = []
codelines = []

if os.path.isfile(os.path.join(os.path.dirname(os.path.realpath(__file__)),
                  "../docs/index.rst")):
    with open(os.path.join(os.path.dirname(os.path.realpath(__file__)),
              "../docs/index.rst")) as f:
        for line in f:
            if line.strip() == ".. code-block:: python":
                code = True
                output = False
                codelines = []
                outputlines = []
            elif code and line.strip() == "::":
                output = True
                code = False
            elif code or output:
                if line.strip() != "":
                    if line.startswith("    "):
                        if code:
                            assert not output
                            codelines.append(line.strip("\n")[4:])
                        else:
                            assert output
                            outputlines.append(line.strip("\n")[4:])
                    else:
                        output = False
                        code = False
                        doc_data.append(("\n".join(codelines), "\n".join(outputlines)))


def test_available_references():
    r_supported = symfem.create_reference.__doc__
    r_supported = r_supported.split("cell_type : str\n")[1]
    r_supported = r_supported.split("vertices : list\n")[0]
    r_supported = r_supported.split("Supported values:")[1]
    r_supported = set([i.strip() for i in r_supported.split(",")])

    e_supported = symfem.create_element.__doc__
    e_supported = e_supported.split("cell_type : str\n")[1]
    e_supported = e_supported.split("element_type : str\n")[0]
    e_supported = e_supported.split("Supported values:")[1]
    e_supported = set([i.strip() for i in e_supported.split(",")])

    assert r_supported == e_supported


def test_available_elements():
    supported = symfem.create_element.__doc__
    supported = supported.split("element_type : str\n")[1]
    supported = supported.split("order : int\n")[0]
    supported = supported.split("Supported values:")[1]
    supported = set([i.strip() for i in supported.split(",")])

    assert set(symfem.create._elementmap.keys()) == supported


@pytest.mark.parametrize("script, output", doc_data)
def test_snippets(script, output):
    root = os.path.join(os.path.dirname(os.path.realpath(__file__)), "..")
    if not os.path.isfile(os.path.join(root, "VERSION")):
        # Skip test if running in tarball source
        pytest.skip()

    old_stdout = sys.stdout
    redirected_output = sys.stdout = StringIO()
    exec(script)
    sys.stdout = old_stdout

    actual_output = redirected_output.getvalue().strip()
    actual_output = re.sub(r"at 0x[^\>]+\>", "at 0x{ADDRESS}>", actual_output)
    assert actual_output == output


def test_version_numbers():
    root = os.path.join(os.path.dirname(os.path.realpath(__file__)), "..")
    if not os.path.isfile(os.path.join(root, "VERSION")):
        # Skip test if running in tarball source
        pytest.skip()

    with open(os.path.join(root, "VERSION")) as f:
        version = f.read()
    assert version == version.strip()

    # codemeta.json
    with open(os.path.join(root, "codemeta.json")) as f:
        data = json.load(f)
    assert data["version"] == version

    # setup.py
    with open(os.path.join(root, "setup.py")) as f:
        for line in f:
            if 'version="' in line:
                assert line.split('version="')[1].split('"')[0] == version

    # symfem/version.py
    with open(os.path.join(root, "symfem/version.py")) as f:
        assert f.read().split('version = "')[1].split('"')[0] == version

    # .github/workflows/test-packages.yml
    url = "https://pypi.io/packages/source/s/symfem/symfem-"
    with open(os.path.join(root, ".github/workflows/test-packages.yml")) as f:
        for line in f:
            if "ref:" in line:
                assert line.split("ref:")[1].strip() == "v" + version
            elif url in line:
                assert line.split(url)[1].split(".tar.gz")[0] == version
            elif "cd symfem-" in line:
                assert line.split("cd symfem-")[1].strip() == version


def test_requirements():
    root = os.path.join(os.path.dirname(os.path.realpath(__file__)), "..")
    if not os.path.isfile(os.path.join(root, "VERSION")):
        # Skip test if running in tarball source
        pytest.skip()

    with open(os.path.join(root, "setup.py")) as f:
        for line in f:
            if 'install_requires=' in line:
                in_setup = [
                    i.strip()[1:-1]
                    for i in line.split('install_requires=[')[1].split(']')[0].split(",")]
    with open(os.path.join(root, "requirements.txt")) as f:
        in_requirements = [i.strip() for i in f]
    assert set(in_requirements) == set(in_setup)


def test_long_description():
    root = os.path.join(os.path.dirname(os.path.realpath(__file__)), "..")
    if not os.path.isfile(os.path.join(root, "VERSION")):
        # Skip test if running in tarball source
        pytest.skip()

    with open(os.path.join(root, "README.md")) as f:
        in_readme = f.read().replace(
            "(logo/logo.png)",
            "(https://raw.githubusercontent.com/mscroggs/symfem/main/logo/logo.png)")

    with open(os.path.join(root, "setup.py")) as f:
        in_setup = f.read().split('long_description = """')[1].split('"""')[0]

    assert in_readme.strip() == in_setup.strip()



from symfem import create_element
import numpy as np
import pytest

elements = {
    "interval": [("P", "Lagrange", range(1, 4)), ("dP", "Discontinuous Lagrange", range(1, 4)),
                 ("serendipity", "Serendipity", range(1, 5)),
                 ("bubble", "Bubble", range(2, 5)),
                 ("dP", "DPC", range(0, 5))],
    "triangle": [("P", "Lagrange", range(1, 4)), ("dP", "Discontinuous Lagrange", range(1, 4)),
                 ("bubble", "Bubble", range(3, 5)),
                 ("N1curl", "Nedelec 1st kind H(curl)", range(1, 4)),
                 ("N2curl", "Nedelec 2nd kind H(curl)", range(1, 4)),
                 ("N1div", "Raviart-Thomas", range(1, 4)),
                 ("N2div", "Brezzi-Douglas-Marini", range(1, 4)),
                 ("Regge", "Regge", range(0, 4)),
                 ("Crouzeix-Raviart", "Crouzeix-Raviart", [1])],
    "tetrahedron": [("P", "Lagrange", range(1, 4)), ("dP", "Discontinuous Lagrange", range(1, 4)),
                    ("bubble", "Bubble", range(4, 6)),
                    ("N1curl", "Nedelec 1st kind H(curl)", range(1, 4)),
                    ("N2curl", "Nedelec 2nd kind H(curl)", range(1, 3)),
                    ("N1div", "Raviart-Thomas", range(1, 3)),
                    ("N2div", "Brezzi-Douglas-Marini", range(1, 3)),
                    ("Regge", "Regge", range(0, 3)),
                    ("Crouzeix-Raviart", "Crouzeix-Raviart", [1])],
    "quadrilateral": [("Q", "Lagrange", range(1, 4)),
                      ("dQ", "Discontinuous Lagrange", range(1, 4)),
                      ("dP", "DPC", range(0, 4)),
                      ("serendipity", "Serendipity", range(1, 5)),
                      ("Qdiv", "Raviart-Thomas", range(1, 4)),
                      ("Qcurl", "Nedelec 1st kind H(curl)", range(1, 4)),
                      ("Sdiv", "Brezzi-Douglas-Marini", range(1, 4)),
                      ("Scurl", "Nedelec 2nd kind H(curl)", range(1, 4))],
    "hexahedron": [("Q", "Lagrange", range(1, 3)), ("dQ", "Discontinuous Lagrange", range(1, 3)),
                   ("serendipity", "Serendipity", range(1, 5)),
                   ("Qdiv", "Raviart-Thomas", range(1, 3)),
                   ("Qcurl", "Nedelec 1st kind H(curl)", range(1, 3)),
                   ("Sdiv", "Brezzi-Douglas-Marini", range(1, 3)),
                   ("Scurl", "Nedelec 2nd kind H(curl)", range(1, 3))],
    "prism": [("Lagrange", "Lagrange", range(1, 4))]
}


def to_float(a):
    try:
        return float(a)
    except:  # noqa: E722
        return [to_float(i) for i in a]


def make_lattice(cell, N=3):
    import numpy as np
    if cell == "interval":
        return np.array([[i / N] for i in range(N + 1)])
    if cell == "triangle":
        return np.array([[i / N, j / N] for i in range(N + 1) for j in range(N + 1 - i)])
    if cell == "tetrahedron":
        return np.array([[i / N, j / N, k / N]
                         for i in range(N + 1) for j in range(N + 1 - i)
                         for k in range(N + 1 - i - j)])
    if cell == "quadrilateral":
        return np.array([[i / N, j / N] for i in range(N + 1) for j in range(N + 1)])
    if cell == "hexahedron":
        return np.array([[i / N, j / N, k / N]
                         for i in range(N + 1) for j in range(N + 1) for k in range(N + 1)])
    if cell == "prism":
        return np.array([[i / N, j / N, k / N]
                         for i in range(N + 1) for j in range(N + 1 - i) for k in range(N + 1)])
    if cell == "pyramid":
        return np.array([[i / N, j / N, k / N]
                         for i in range(N + 1) for j in range(N + 1)
                         for k in range(N + 1 - max(i, j))])


@pytest.mark.parametrize(("cell", "symfem_type", "basix_type", "order"),
                         [(cell, a, b, order) for cell, ls in elements.items()
                          for a, b, c in ls for order in c])
def test_against_basix(has_basix, elements_to_test, cells_to_test, cell, symfem_type,
                       basix_type, order, speed):
    if elements_to_test != "ALL" and symfem_type not in elements_to_test:
        pytest.skip()
    if cells_to_test != "ALL" and cell not in cells_to_test:
        pytest.skip()
    if speed == "fast" and order > 2:
        pytest.skip()

    if has_basix:
        import basix
    else:
        try:
            import basix
        except ImportError:
            pytest.skip("Basix must be installed to run this test.")

    points = make_lattice(cell, 2)
    space = basix.create_element(basix_type, cell, order)
    result = space.tabulate(0, points)[0]

    element = create_element(cell, symfem_type, order)
    sym_result = element.tabulate_basis(points, "xxyyzz", symbolic=False, use_legendre=True)

    assert np.allclose(result, sym_result)



import sys
import setuptools

if sys.version_info < (3, 6):
    print("Python 3.6 or higher required, please upgrade.")
    sys.exit(1)

long_description = """
![Symfem](https://raw.githubusercontent.com/mscroggs/symfem/main/logo/logo.png)

[![Documentation status](https://readthedocs.org/projects/symfem/badge/?version=latest)](https://symfem.readthedocs.io/en/latest/?badge=latest)
[![Style checks](https://github.com/mscroggs/symfem/actions/workflows/style-checks.yml/badge.svg)](https://github.com/mscroggs/symfem/actions)
[![Run tests](https://github.com/mscroggs/symfem/actions/workflows/run-tests.yml/badge.svg)](https://github.com/mscroggs/symfem/actions)
[![Coverage Status](https://coveralls.io/repos/github/mscroggs/symfem/badge.svg?branch=main)](https://coveralls.io/github/mscroggs/symfem?branch=main)
[![PyPI](https://img.shields.io/pypi/v/symfem?color=blue&label=PyPI&logo=pypi&logoColor=white)](https://pypi.org/project/symfem/)

Symfem is a symbolic finite element definition library, that can be used to
symbolically evaluate the basis functions of a finite element space.

## Installing Symfem
### Installing from repo
Symfem can be installed by downloading the [GitHub repo](https://github.com/mscroggs/symfem)
and running:

```bash
python3 setup.py install
```

### Installing using pip
The latest release of Symfem can be installed by running:

```bash
pip3 install symfem
```

### Installing using conda
The latest release of Symfem can be installed by running:

```bash
conda install symfem
```

## Testing Symfem
To run the Symfem unit tests, clone the repository and run:

```bash
python3 -m pytest test/
```

## Using Symfem
Documentation of the latest release version of Symfem can be found on
[Read the Docs](https://symfem.readthedocs.io/en/latest/).

## Contributing to Symfem
You can find information about how to contribute to Symfem [here](CONTRIBUTING.md).
"""

data_files = ["LICENSE", "requirements.txt",
              ("test", ["test/__init__.py", "test/utils.py", "test/conftest.py"])]

if __name__ == "__main__":
    setuptools.setup(
        name="symfem",
        description="a symbolic finite element definition library",
        long_description=long_description,
        long_description_content_type="text/markdown",
        version="2021.7.6",
        author="Matthew Scroggs",
        license="MIT",
        author_email="symfem@mscroggs.co.uk",
        maintainer_email="symfem@mscroggs.co.uk",
        url="https://github.com/mscroggs/symfem",
        packages=["symfem", "symfem.elements"],
        include_package_data=True,
        data_files=data_files,
        install_requires=["sympy", "numpy"]
    )



"""Symbolic tools."""

import sympy


def to_sympy(i):
    """Convert to a sympy expression."""
    if isinstance(i, list):
        return [to_sympy(j) for j in i]
    if isinstance(i, tuple):
        return tuple(to_sympy(j) for j in i)

    if isinstance(i, int):
        return sympy.Integer(i)
    if isinstance(i, Monomial):
        return i.to_sympy()

    return i


def to_float(i):
    """Convert to a float."""
    if isinstance(i, list):
        return [to_float(j) for j in i]
    if isinstance(i, tuple):
        return tuple(to_float(j) for j in i)

    return float(i)


class Monomial:
    """A monomial."""

    def __init__(self, xpow=0, ypow=0, zpow=0, negative=False):
        self._x = xpow
        self._y = ypow
        self._z = zpow
        self._negative = negative

    @property
    def order(self):
        """Get the order of the monomial."""
        return self._x + self._y + self._z

    @property
    def indices(self):
        """Get the indices of the monomial."""
        return (self._x, self._y, self._z)

    def diff(self, variable):
        """Differentiate the monomial."""
        return self.to_sympy().diff(to_sympy(variable))

    def to_sympy(self):
        """Convert to a sympy expression."""
        _x = [sympy.Symbol("x"), sympy.Symbol("y"), sympy.Symbol("z")]

        if self._x + self._y + self._z == 1 and not self._negative:
            if self._x == 1:
                return _x[0]
            if self._y == 1:
                return _x[1]
            if self._z == 1:
                return _x[2]

        if self._negative:
            return -_x[0] ** self._x * _x[1] ** self._y * _x[2] ** self._z
        else:
            return _x[0] ** self._x * _x[1] ** self._y * _x[2] ** self._z

    def __hash__(self):
        """Return hash."""
        return hash(self.to_sympy())

    def __unicode__(self):
        """Return unicode."""
        return str(self.to_sympy())

    def __str__(self):
        """Return string."""
        return self.__unicode__()

    def __repr__(self):
        """Return representation."""
        return self.__unicode__()

    def __eq__(self, other):
        """Check if monomials are equal."""
        if isinstance(other, Monomial):
            return (self._x == other._x and self._y == other._y
                    and self._z == other._z and self._negative == other._negative)
        return self.to_sympy() == other

    def __mul__(self, other):
        """Multiply."""
        if isinstance(other, Monomial):
            return Monomial(self._x + other._x, self._y + other._y,
                            self._z + other._z, self._negative)
        return self.to_sympy() * other

    def __rmul__(self, other):
        """Multiply."""
        return self.__mul__(other)

    def __truediv__(self, other):
        """Divide."""
        if isinstance(other, Monomial):
            return Monomial(self._x - other._x, self._y - other._y,
                            self._z - other._z, self._negative)
        return self.to_sympy() / to_sympy(other)

    def __rtruediv__(self, other):
        """Divide."""
        if isinstance(other, Monomial):
            return Monomial(other._x - self._x, other._y - self._y,
                            other._z - self._z, self._negative)
        return other / self.to_sympy()

    def __pow__(self, power):
        """Exponentiate."""
        return Monomial(self._x * power, self._y * power, self._z * power, self._negative)

    def __neg__(self):
        """Negate."""
        return Monomial(self._x, self._y, self._z, not self._negative)

    def __add__(self, other):
        """Add."""
        return self.to_sympy() + other

    def __radd__(self, other):
        """Add."""
        return self.__add__(other)

    def __sub__(self, other):
        """Subtract."""
        return self.to_sympy() - other

    def __rsub__(self, other):
        """Subtract."""
        return other - self.to_sympy()

    def __getattr__(self, attr):
        """Forward all other calls to sympy."""
        return getattr(self.to_sympy(), attr)


x = [Monomial(xpow=1), Monomial(ypow=1), Monomial(zpow=1)]

t = [sympy.Symbol("t0"), sympy.Symbol("t1"), sympy.Symbol("t2")]

_dummy = [sympy.Symbol("symbolicpyDUMMYx"), sympy.Symbol("symbolicpyDUMMYy"),
          sympy.Symbol("symbolicpyDUMMYz")]


def subs(f, vars, values):
    """Substitute values into a sympy expression."""
    if isinstance(f, Monomial):
        return subs(f.to_sympy(), vars, values)
    if isinstance(vars, Monomial):
        return subs(f, vars.to_sympy(), values)
    elif isinstance(vars, (list, tuple)):
        for i, j in enumerate(vars):
            if isinstance(j, Monomial):
                return subs(f, vars[:i] + [j.to_sympy()] + vars[i + 1:], values)
    if isinstance(values, Monomial):
        return subs(f, vars, values.to_sympy())
    elif isinstance(values, (list, tuple)):
        for i, j in enumerate(values):
            if isinstance(j, Monomial):
                return subs(f, vars, values[:i] + [j.to_sympy()] + values[i + 1:])

    if isinstance(f, PiecewiseFunction):
        return f.evaluate(values)
    try:
        return tuple(subs(f_j, vars, values) for f_j in f)
    except TypeError:
        pass
    if isinstance(vars, sympy.Symbol):
        return to_sympy(f).subs(vars, values)

    if isinstance(f, int):
        return f

    if len(values) == 1:
        return f.subs(vars[0], values[0])
    if len(values) == 2:
        return f.subs(vars[0], _dummy[0]).subs(vars[1], _dummy[1]).subs(
            _dummy[0], values[0]).subs(_dummy[1], values[1])
    if len(values) == 3:
        return f.subs(vars[0], _dummy[0]).subs(vars[1], _dummy[1]).subs(
            vars[2], _dummy[2]).subs(_dummy[0], values[0]).subs(
                _dummy[1], values[1]).subs(_dummy[2], values[2])


def sym_sum(ls):
    """Symbolically computes the sum of a list."""
    out = to_sympy(0)
    for i in ls:
        out += i
    return out


class PiecewiseFunction:
    """A function defined piecewise on a collection of triangles."""

    def __init__(self, pieces):
        self.pieces = pieces

    def get_piece(self, point):
        """Get the piece of the function defined at the given point."""
        from .vectors import point_in_triangle
        for tri, value in self.pieces:
            if point_in_triangle(point[:2], tri):
                return value

        raise NotImplementedError("Evaluation of piecewise functions outside domain not supported.")

    def evaluate(self, values):
        """Evaluate a function."""
        try:
            return subs(self.get_piece(values), x, values)
        except TypeError:
            return PiecewiseFunction([(i, subs(j, x, values)) for i, j in self.pieces])

    def diff(self, variable):
        """Differentiate the function."""
        from .calculus import diff
        return PiecewiseFunction([(i, diff(j, variable)) for i, j in self.pieces])

    def __rmul__(self, other):
        """Multiply the function by a scalar."""
        return PiecewiseFunction([(i, other * j) for i, j in self.pieces])

    def __mul__(self, other):
        """Multiply the function by a scalar."""
        return self.__rmul__(other)

    def __radd__(self, other):
        """Add another piecewise function or a scalar."""
        return self.__add__(other)

    def __add__(self, other):
        """Add another piecewise function or a scalar."""
        if isinstance(other, PiecewiseFunction):
            for i, j in zip(self.pieces, other.pieces):
                assert i[0] == j[0]
            return PiecewiseFunction(
                [(i[0], i[1] + j[1]) for i, j in zip(self.pieces, other.pieces)])

        return PiecewiseFunction([(i, other + j) for i, j in self.pieces])
